# coding: utf8

"""
This module includes the `Show` class and the main functionalities of `auditorium`.
"""

import base64
import io
import os
import webbrowser
from collections import OrderedDict

from jinja2 import Template
from markdown import markdown
from sanic import Sanic
from sanic.response import html, json


from .components import ShowMode, Animation, Column, Vertical, Fragment, Block
from .utils import fix_indent


def path(p):
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), p)


class Show:
    def __init__(self, title="", theme='white'):
        self.slides = {}
        self.slide_ids = []
        self.vertical_slides = {}
        self.theme = theme

        self.app = Sanic("auditorium")
        self.app.route("/")(self._index)
        self.app.route("/update", methods=['POST'])(self._update)
        self.app.static('static', path('static'))

        self.current_content = []
        self.current_update = {}
        self.current_slide = None
        self._title = title
        self._unique_id = 0
        self._global_values = {}
        self._mode = ShowMode.Edit

        with open(path('templates/index.html')) as fp:
            self._template = Template(fp.read())

    ## Show functions

    def run(self, host, port, launch, *args, **kwargs):
        self._html = self.render()

        if launch:
            def launch_server():
                webbrowser.open_new_tab(f"http://{host}:{port}")

            self.app.add_task(launch_server)

        self.app.run(host=host, port=port, *args, **kwargs)

    @property
    def show_title(self):
        return self._title

    ## @slide decorator

    def slide(self, func=None, id=None):
        if func is not None:
            return self._wrap(func, id)

        elif id is not None:
            def wrapper(func):
                return self._wrap(func, id)

            return wrapper

    def _wrap(self, func, id):
        slide_id = id or func.__name__

        if self._mode == ShowMode.Edit:
            self.slide_ids.append(slide_id)
            self.vertical_slides[slide_id] = OrderedDict()
        elif self._mode == ShowMode.Markup:
            self.vertical_slides[self.current_slide][slide_id] = None
        else:
            return

        self.slides[slide_id] = func
        return func

    ## Binding methods

    def title(self, text):
        return self.header(text, 1)

    def header(self, text, level=2):
        return self.markdown("#" * level + " " + text)

    def hrule(self):
        return self.markup("<hr>")

    def markup(self, content):
        item_id, id_markup = self._get_unique_id("markup")

        if self._mode == ShowMode.Markup:
            self.current_content.append(f'<div {id_markup}>{fix_indent(content)}</div>')
        else:
            self.current_update[item_id] = fix_indent(content)

    def markdown(self, content):
        item_id, id_markup = self._get_unique_id("markdown")

        if self._mode == ShowMode.Markup:
            self.current_content.append(f'<div {id_markup}>{markdown(fix_indent(content))}</div>')
        else:
            self.current_update[item_id] = markdown(fix_indent(content))

    def text_input(self, default="", track_keys=True):
        item_id, id_markup = self._get_unique_id("text-input")

        event = "onkeyup" if track_keys else "onchange"

        if self._mode == ShowMode.Markup:
            self.current_content.append(f'<input {id_markup} data-event="{event}" type="text" class="text" value="{default}"></input>')
            self._global_values[item_id] = default
            return default
        else:
            return self._global_values[item_id]

    def pyplot(self, plt, height=400, aspect=4/3, fmt="png"):
        item_id, id_markup = self._get_unique_id("pyplot")
        plt.tight_layout()
        buffer = io.BytesIO()
        plt.savefig(buffer, format=fmt)
        plt.clf()
        buffer.seek(0)

        if fmt == 'png':
            code = base64.b64encode(buffer.read(-1)).decode('ascii')
            markup = f'<img width="{aspect * height}px" height="{height}px" class="pyplot" src="data:image/png;base64, {code}"></img>'
        elif fmt == 'svg':
            code = buffer.read(-1).decode('ascii')
            markup = code
        else:
            raise ValueError("Invalid value for parameter `fmt`, must be 'png' or 'svg'.")

        if self._mode == ShowMode.Markup:
            self.current_content.append(f"<div {id_markup}>{markup}</div>")
        else:
            self.current_update[item_id] = markup

    def anchor(self, slide_or_href, text=None):
        item_id, id_markup = self._get_unique_id("anchor")

        if self._mode == ShowMode.Markup:
            if hasattr(slide_or_href, '__name__'):
                slide_or_href = "#/" + slide_or_href.__name__

            if text is None:
                text = slide_or_href

            self.current_content.append(f'<a {id_markup} href="{slide_or_href}">{text}</a>')
        else:
            self.current_update[item_id] = text

    def code(self, text, language='python'):
        _, id_markup = self._get_unique_id("code")
        content = fix_indent(text)

        if self._mode == ShowMode.Markup:
            self.current_content.append(f'<div {id_markup}><pre><code class="{language}">{content}</pre></code></div>')
        # else:
        #     self.current_update[item_id] = markdown(content)

    def animation(self, steps=10, time=0.1, loop=True):
        item_id, id_markup = self._get_unique_id('animation')

        if self._mode == ShowMode.Markup:
            self.current_content.append(f'<animation {id_markup} data-steps="{steps}" data-time="{time}" data-loop="{loop}"></animation>')
            self._global_values[item_id] = Animation(steps, time, loop)

        return self._global_values[item_id]

    def columns(self, *widths):
        return Column(self, *widths)

    def vertical(self):
        return Vertical(self)

    def fragment(self, style='fade-in'):
        return Fragment(self, style)

    def block(self, title="", style='default'):
        return Block(self, title, style)

    def success(self, title=""):
        return self.block(title, 'success')

    def warning(self, title=""):
        return self.block(title, 'warning')

    def error(self, title=""):
        return self.block(title, 'error')

    ## Internal API

    def do_markup(self, slide):
        self._mode = ShowMode.Markup
        self.current_content.clear()
        self._run(slide)
        self._mode = ShowMode.Edit
        return "\n\n".join(self.current_content)

    def do_code(self, slide):
        self._mode = ShowMode.Code
        self.current_update.clear()
        self._run(slide)
        self._mode = ShowMode.Edit
        return self.current_update

    ## Utilities

    def _run(self, slide):
        self._unique_id = 0
        self.current_slide = slide
        func = self.slides[slide]

        if func.__doc__:
            self.markdown(func.__doc__)

        func()

    def _get_unique_id(self, markup):
        self._unique_id += 1
        item_id = f"{self.current_slide}-{markup}-{self._unique_id - 1}"
        return item_id, f'id="{item_id}" data-slide="{self.current_slide}"'

    def render(self, theme=None):
        return self._template.render(show=self, theme=theme or self.theme)

    ## Routes

    async def _update(self, request):
        data = request.json

        if data['type'] == 'input':
            self._global_values[data['id']] = data['value']
        elif data['type'] == 'animation':
            self._global_values[data['id']].next()

        self.do_code(data['slide'])
        return json(self.current_update)

    async def _index(self, request):
        theme = request.args.get("theme", self.theme)

        if theme != self.theme:
            return html(self.render(theme))

        return html(self._html)



# coding: utf8

import fire
import runpy
import webbrowser

from auditorium.markdown import MarkdownLoader


class Auditorium:
    @staticmethod
    def run(path, host='localhost', port=6789, debug=False, instance_name='show', launch=True):
        "Runs a custom Python script as a slideshow."

        if path.endswith('.py'):
            ns = runpy.run_path(path)
            show = ns[instance_name]
        elif path.endswith('.md'):
            loader = MarkdownLoader(path, instance_name=instance_name)
            show = loader.parse()

        show.run(host=host, port=port, debug=debug, launch=launch)

    @staticmethod
    def demo(host='localhost', port=6789, debug=False, launch=True):
        "Starts the demo slideshow."

        from auditorium.demo import show
        show.run(host, port, debug=debug, launch=launch)


def main():
    fire.Fire(Auditorium, name='auditorium')


if __name__ == "__main__":
    main()



# coding: utf8


def fix_indent(content, tab_size=0):
    lines = content.split("\n")
    min_indent = 1000

    for l in lines:
        if not l or l.isspace():
            continue

        indent_size = 0

        for c in l:
            if c.isspace():
                indent_size += 1
            else:
                break

        min_indent = min(indent_size, min_indent)

    if min_indent < 10000:
        lines = [" " * tab_size + l[min_indent:] for l in lines]

    while lines and (not lines[0] or lines[0].isspace()):
        lines.pop(0)

    return "\n".join(lines)



# coding: utf8

from auditorium import Show


class MarkdownLoader:
    def __init__(self, path, instance_name='show'):
        self.path = path
        self.instance_name = instance_name

    def parse(self):
        slides = []
        current_slide = []

        with open(self.path) as fp:
            for line in fp:
                line = line.strip("\n")

                if line.startswith("## ") and current_slide:
                    slides.append(current_slide)
                    current_slide = []

                current_slide.append(line)

            if current_slide:
                slides.append(current_slide)

        show = Show()

        for i, slide in enumerate(slides):
            show.slide(func=MarkdownSlide(show, slide), id='slide-%i' % (i+1))

        return show


class MarkdownSlide:
    def __init__(self, show: Show, content):
        self.show = show
        self.content = []

        state = 'markdown' # or 'code'
        language = ''
        code_start = ''
        split = []

        for line in content:
            if state == 'markdown':
                if line.startswith('```') or line.startswith('~~~'):
                    if split:
                        self.content.append(MarkdownContent(split))

                    split = []
                    state = 'code'
                    code_start = line[:3]
                    language = line[3:].split()[0]
                    tags = line[3:].split()[1:]
                else:
                    split.append(line)

            elif state == 'code':
                if line == code_start:
                    if split:
                        self.content.append(PythonContent(split, language, tags))

                    split = []
                    state = 'markdown'
                else:
                    split.append(line)

        if split:
            if state == 'markdown':
                self.content.append(MarkdownContent(split))
            else:
                raise ValueError("Didn't closed a Python line...")

    def __call__(self):
        global_context = dict(show=self.show)

        for content in self.content:
            content(self.show, global_context)


class MarkdownContent:
    def __init__(self, lines):
        self.lines = "\n".join(lines)

    def __call__(self, show, global_context):
        show.markdown(self.lines.format(**global_context))


class PythonContent:
    def __init__(self, lines, language, tags):
        self.lines = "\n".join(lines)
        self.tags = tags
        self.language = language

    def __call__(self, show, global_context):
        run = ':run' in self.tags
        echo = not run or ':echo' in self.tags
        persist = ':persist' in self.tags

        local_context = dict()

        if run:
            exec(self.lines, global_context, local_context)

        if echo:
            show.code(self.lines, self.language)

        if persist:
            global_context.update(local_context)



# coding: utf8

"""
A simple demo script for `auditorium`.
"""

from auditorium import Show

show = Show("Auditorium Demo")


@show.slide
def intro():
    """
    # Auditorium
    ## A Python-powered slideshow generator with steroids
    """


@show.slide
def what_is_this():
    """
    ## What is Auditorium

    `auditorium` is a Python module for creating slideshows which
    are ultimately displayed in a browser using the amazing
    library `reveal.js`.

    With `auditorium` you don't need to learn JavaScript,
    HTML or CSS. Everything goes in your Python code,
    both presentation content and logic.
    """


@show.slide
def how_it_works():
    """
    ## How it works

    `auditorium` creates a continuous feedback loop between Python
    and HTML/JavaScript.

    * Every slide in `auditorium` is a Python method.
    * You can mix up content with logic.
        - Inject variables into the presentation.
        - Receive back interaction events and values.
    * Layout is handled by `reveal.js`.
    * You only ever need to write Python code.
    """


@show.slide
def the_basics():
    show.markdown("## The basics")
    show.markdown("Start by declaring a `Show` instance.")

    show.code("""
    from auditorium import Show
    show = Show("Auditorium Demo")""")

    show.markdown("Add a method for every slide, decorated with `@show.slide`.")

    show.code("""
    @show.slide
    def my_slide():
        # slide content
    """)

    show.markdown("Finally run the show.")

    show.code("auditorium run my_show.py", 'bash')


@show.slide
def examples():
    show.header("Some examples")
    show.markdown("Slide content is added with _markdown_.")
    show.markup("Or <strong>directly</strong> as HTML.")

    show.hrule()

    show.code("""
    @show.slide
    def example():
        show.header("Some examples")
        show.markdown("Slide content is added with _markdown_.")
        show.markup("Or <strong>directly</strong> as HTML.")
    """)


@show.slide
def static_content():
    """
    ## Static Content

    Static content can also be added directly as markdown in
    the `docstring` of the corresponding method.
    """

    show.hrule()

    show.code("""
    @show.slide
    def static_content():
        \"\"\"
        ## Static Content

        Static content can also be added directly as markdown in
        the `docstring` of the corresponding method.
        \"\"\"
    """)


@show.slide
def layout():
    """
    ## Custom layout

    More complex layouts can be composed.
    """

    show.hrule()

    with show.columns(2, 3) as cl:
        show.markdown("#### Content")
        show.markdown("You can put custom content in any column...")

        cl.tab()
        show.code("""
        with show.columns(2, 3) as cl:
            show.markdown("#### Content")
            show.markdown("...")

            cl.tab()
            show.code("...")
        """)


@show.slide
def data_binding():
    """
    ## Dynamic Data

    Use dynamic data with two-way binding to add executable Python logic
    to your presentation.
    """

    with show.columns(2) as cl:
        text = show.text_input("dlrow")
        text = "".join(reversed(text)).title()
        cl.tab()
        show.markdown(f"Hello {text}!!")

    show.hrule()

    show.code("""
    @show.slide
    def data_binding():
        # ...
        text = show.text_input("dlrow")
        text = "".join(reversed(text)).title()
        show.markdown(f"Hello {text}!!")
    """)


@show.slide
def animation():
    """
    ## Animations

    You can create simple stateless animations with pure Python
    """

    with show.animation(steps=10, time=0.33, loop=True) as anim:
        show.markdown("> ### ." + ("." * anim.current))

    show.hrule()

    show.code("""
    with show.animation(steps=10, time=0.33, loop=True) as anim:
        show.markdown("> ### ." + ("." * anim.current))
    """)

@show.slide
def vertical_slides():
    """
    ## Vertical Slides

    Vertical slides allow you to create "read-more" like
    features in your slideshow.
    They can be skipped on shorter presentations and left
    for more interested audiences.
    """

    with show.warning():
        show.markdown('Press `DOWN` instead of `LEFT` or click the down arrow.')

    @show.slide
    def vertical_code():
        """
        ## Vertical Slides: Code
        """

        show.code("""
        @show.slide
        def main_slide():
            # content of main slide

            @show.slide
            def vertical_1():
                # content of first vertical slide

            @show.slide
            def vertical_2():
                # content of first vertical slide
        """)

    @show.slide
    def vertical_more():
        """
        ### Vertical Slides: More Info

        If you press `SPACE` instead of `LEFT` or `DOWN`
        the slideshow will cycle through **all** of the slides
        in order, including vertical slides.

        It will go down and left automatically.
        """

@show.slide
def blocks():
    """
    ## Blocks

    Like beamer, pre-styled blocks are available.
    """

    with show.columns(2) as cl:

        with show.block('Standard block'):
            show.markdown("And its content...")

        with show.success('Success block'):
            show.markdown("For happy endings...")

        cl.tab()

        with show.warning('Warning block'):
            show.markdown("For hairy stuff...")

        with show.error('Error block'):
            show.markdown("When nothing works...")

    @show.slide
    def blocks_code():
        """
        ## Blocks: Code
        """

        show.code("""
        with show.block('Standard block'):
            show.markdown("And its content...")

        with show.success('Success block'):
            show.markdown("For happy endings...")

        with show.warning('Warning block'):
            show.markdown("For hairy stuff...")

        with show.error('Error block'):
            show.markdown("When nothing works...")
        """)


@show.slide
def fragments():
    """
    ## Fragments

    Fragments allow to animate elements inside a slide.
    """

    with show.fragment():
        show.markdown("The can have different animations as well...")

    with show.fragment():
        show.hrule()
        show.code("""
        with show.fragment(style='...'): # fade-in, grow, ...
            # content
        """)

    @show.slide
    def fragment_examples():
        """
        ## Fragment examples

        Here are some of the possible fragment animations.
        """

        with show.columns(3) as cl:
            for i, style in enumerate('grow shrink fade-in fade-out fade-up fade-down fade-left fade-right highlight-blue highlight-red highlight-green'.split()):
                if i > 0 and i % 4 == 0:
                    cl.tab()

                with show.fragment(style):
                    show.markdown(f'`{style}`')


@show.slide
def pyplot():
    """
    ## pyplot

    Dynamically generated graphs with `pyplot` can be added
    also very easily.
    """
    from matplotlib import pyplot as plt
    import numpy as np

    xg = np.random.RandomState(0)
    yg = np.random.RandomState(1)

    with show.columns(2) as cl:
        with show.animation(steps=60, time=0.5, loop=True) as anim:
            x = xg.uniform(size=anim.current * 50)
            y = yg.uniform(size=anim.current * 50)
            colors = ['green' if xi ** 2 + yi ** 2 < 1 else 'orange' for (xi, yi) in zip(x,y)]
            plt.scatter(x, y, s=3, c=colors)
            plt.ylim(0, 1)
            plt.xlim(0, 1)
            show.pyplot(plt, fmt='png', height=350)

            cl.tab()

            with show.block('Monte Carlo Sampling'):
                show.markdown(f"Samples: {len(x)}")
                show.markdown(f"Inside: {colors.count('green')}")
                show.markdown("**Pi = %.3f**" % (4 * colors.count('green') / len(x) if len(x) > 0 else 0))

    @show.slide
    def pyplot_code():
        """### Pyplot: Code"""

        show.code("""
        xg = np.random.RandomState(0)
        yg = np.random.RandomState(1)

        with show.animation(steps=60, time=0.5, loop=True) as anim:
            x = xg.uniform(size=anim.current * 50)
            y = yg.uniform(size=anim.current * 50)
            colors = ['green' if xi ** 2 + yi ** 2 < 1 else 'orange'
                       for (xi, yi) in zip(x,y)]
            plt.scatter(x, y, s=3, c=colors)
            plt.ylim(0, 1)
            plt.xlim(0, 1)
            show.pyplot(plt, fmt='png', height=350)
        """)


@show.slide
def themes():
    """
    ## Themes
    """

    show.markdown("A standard `reveal.js` theme can be specified at creation time:")
    show.code("show = Show(theme='black')")

    show.markdown("Or directly as a query string arg:")

    with show.block():
        show.anchor("/?theme=black#/themes")



# coding: utf8

from enum import IntEnum


class ShowMode(IntEnum):
    Edit = 0
    Markup = 1
    Code = 2


class Animation:
    def __init__(self, steps, time, loop):
        self.steps = steps
        self.time = time
        self.loop = loop
        self._current = 0

    @property
    def current(self):
        return self._current

    def next(self):
        self._current += 1

        if self._current >= self.steps:
            if self.loop:
                self._current = 0
            else:
                self._current = self.steps - 1

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass


class Wrapper:
    def __init__(self, show):
        self.show = show
        self.begin()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        return self.end()

    def begin(self):
        raise NotImplementedError()

    def end(self):
        raise NotImplementedError()


class Column(Wrapper):
    def __init__(self, show, *widths):
        if len(widths) == 1:
            widths = [1.0/widths[0] for _ in range(widths[0])]

        total_width = sum(widths) + 0.01
        widths = [w/total_width for w in widths]

        self.widths = list(widths)
        super(Column, self).__init__(show)

    def begin(self):
        self.show.current_content.append(f'<div class="columns">')
        self.show.current_content.append(f'<div class="column" style="width: {self.widths[0] * 100}%;">')
        self.widths.pop(0)

    def tab(self):
        self.show.current_content.append(f'</div>')
        self.show.current_content.append(f'<div class="column" style="width: {self.widths[0] * 100}%;">')
        self.widths.pop(0)

    def end(self):
        self.show.current_content.append('</div>')
        self.show.current_content.append('</div>')


class Vertical(Wrapper):
    def begin(self):
        self.show.slide_ids.append("show::start-section")

    def end(self):
        self.show.slide_ids.append("show::end-section")


class Block(Wrapper):
    def __init__(self, show, title, style):
        self.title = title
        self.style = style
        super(Block, self).__init__(show)

    def begin(self):
        self.show.current_content.append(f'<div class="block block-{self.style}"><h1 class="block-title">{self.title}</h1><div class="block-content">')

    def end(self):
        self.show.current_content.append('</div></div>')


class Fragment(Wrapper):
    def __init__(self, show, style):
        self.style = style
        super(Fragment, self).__init__(show)

    def begin(self):
        self.show.current_content.append(f'<div class="fragment {self.style}">')

    def end(self, *args):
        self.show.current_content.append('</div>')



# coding: utf8

import toml
from setuptools import setup


# TODO: Update version whenever changes
VERSION = '0.4.5'


def get_install_requirements():
    """Automatically pull requirements from Pipfile.
    Adapted from: <https://medium.com/homeaway-tech-blog/simplifying-python-builds-74e76802444f>
    """
    try:
        # read my pipfile
        with open ('Pipfile', 'r') as fh:
            pipfile = fh.read()
        # parse the toml
        pipfile_toml = toml.loads(pipfile)
    except FileNotFoundError:
        return []
    # if the package's key isn't there then just return an empty
    # list
    try:
        required_packages = pipfile_toml['packages'].items()
    except KeyError:
        return []
    # If a version/range is specified in the Pipfile honor it
    # otherwise just list the package
    return ["{0}{1}".format(pkg,ver) if ver != "*"
            else pkg for pkg,ver in required_packages]


setup(
    # TODO: Change your library name and additional information down here
    name='auditorium',
    packages=['auditorium'],
    url='https://github.com/apiad/auditorium',
    download_url='https://github.com/apiad/auditorium/tarball/{}'.format(VERSION),
    license='MIT',
    author='Alejandro Piad',
    author_email='apiad@apiad.net',
    description='A Python-powered slideshow maker with steroids.',
    include_package_data=True,
    zip_safe=False,

    # This should automatically take your long description from Readme.md
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',

    # This should automatically pull your requirements from `Pipfile`
    install_requires=get_install_requirements(),
    version=VERSION,

    # TODO (Optional): Set your entry-points (CLI apps to register) here
    entry_points={
        'console_scripts': ['auditorium=auditorium.__main__:main'],
    },

    # TODO: Choose your classifiers carefully
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)
